import argparse
import shlex
import tempfile
from pathlib import Path
from textwrap import dedent
from textwrap import indent
from typing import Any
from typing import Callable
from typing import cast

from IPython.core.error import UsageError
from IPython.core.magic import Magics
from IPython.core.magic import cell_magic
from IPython.core.magic import magics_class
from IPython.display import IFrame
from IPython.display import display
from rich import print as pprint

from memray import FileReader
from memray.commands.common import warn_if_not_enough_symbols
from memray.reporters.flamegraph import FlameGraphReporter

_typed_cell_magic = cast(
    Callable[[Callable[..., Any]], Callable[..., Any]],
    cell_magic,
)

TEMPLATE = """\
from memray import Tracker, FileFormat
with Tracker(
    "{dump_file!s}",
    native_traces={native_traces},
    trace_python_allocators={trace_python_allocators},
    follow_fork={follow_fork},
    file_format=FileFormat.{file_format},
) as tracker:
{code}
"""


def _run_cell(shell: Any, code: str) -> None:
    run_cell = cast(Callable[[str], Any], shell.run_cell)
    run_cell(code)


def _display_iframe(frame: IFrame) -> None:
    show = cast(Callable[[IFrame], Any], display)
    show(frame)


def argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="%%memray_flamegraph")
    parser.add_argument(
        "--native",
        help="Track native (C/C++) stack frames as well",
        action="store_true",
        dest="native",
        default=False,
    )
    parser.add_argument(
        "--follow-fork",
        action="store_true",
        help="Record allocations in child processes forked from the tracked script",
        default=False,
    )
    parser.add_argument(
        "--trace-python-allocators",
        action="store_true",
        help="Record allocations made by the pymalloc allocator",
        default=False,
    )
    alloc_type_group = parser.add_mutually_exclusive_group()
    alloc_type_group.add_argument(
        "--leaks",
        help="Show memory leaks, instead of peak memory usage",
        action="store_true",
        dest="show_memory_leaks",
        default=False,
    )
    alloc_type_group.add_argument(
        "--temporary-allocation-threshold",
        metavar="N",
        help=dedent(
            """
            Report temporary allocations, as opposed to leaked allocations
            or high watermark allocations.  An allocation is considered
            temporary if at most N other allocations occur before it is
            deallocated.  With N=0, an allocation is temporary only if it
            is immediately deallocated before any other allocation occurs.
            """
        ),
        action="store",
        dest="temporary_allocation_threshold",
        type=int,
        default=-1,
    )
    alloc_type_group.add_argument(
        "--temporary-allocations",
        help="Equivalent to --temporary-allocation-threshold=1",
        action="store_const",
        dest="temporary_allocation_threshold",
        const=1,
    )

    parser.add_argument(
        "--temporal",
        help=(
            "Generate a dynamic flame graph that can analyze"
            " allocations in a user-selected time range."
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--split-threads",
        help="Do not merge allocations across threads",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--inverted",
        help=(
            "Invert the flame graph: "
            "use allocators as roots instead of thread entry points"
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--max-memory-records",
        help="Maximum number of memory records to display",
        type=int,
        default=None,
    )

    return parser


@magics_class
class FlamegraphMagics(Magics):
    @_typed_cell_magic
    def memray_flamegraph(self, line: str, cell: str) -> None:
        """Memory profile the code in the cell and display a flame graph."""
        if self.shell is None:
            raise UsageError("Cannot profile code when not in a shell")

        try:
            options = argument_parser().parse_args(shlex.split(line))
        except SystemExit:
            # argparse wants to bail if the options aren't valid.
            # It already printed a message, just return control to IPython.
            return

        if options.temporal and options.temporary_allocation_threshold >= 0:
            raise UsageError(
                "Can't create a temporal flame graph of temporary allocations"
            )
        elif options.temporal or options.temporary_allocation_threshold >= 0:
            file_format = "ALL_ALLOCATIONS"
        else:
            file_format = "AGGREGATED_ALLOCATIONS"

        results_dir = Path("memray-results")
        results_dir.mkdir(exist_ok=True)

        tempdir = Path(tempfile.mkdtemp(dir=results_dir))
        if tempdir.is_absolute():
            tempdir = tempdir.relative_to(Path.cwd())
        dump_file = Path(tempdir) / "memray.dump"
        code = TEMPLATE.format(
            dump_file=dump_file,
            native_traces=options.native,
            trace_python_allocators=options.trace_python_allocators,
            follow_fork=options.follow_fork,
            code=indent(cell, " " * 4),
            file_format=file_format,
        )
        _run_cell(self.shell, code)

        merge_threads = not options.split_threads

        reporter = None

        kwargs = {}
        if options.max_memory_records is not None:
            kwargs["max_memory_records"] = options.max_memory_records

        with FileReader(dump_file, report_progress=True, **kwargs) as reader:
            if reader.metadata.has_native_traces:
                warn_if_not_enough_symbols()

            if options.temporal:
                if options.show_memory_leaks:
                    temporal_snapshot = reader.get_temporal_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = FlameGraphReporter.from_temporal_snapshot(
                        temporal_snapshot,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=None,
                        inverted=options.inverted,
                    )
                else:
                    recs, hwms = reader.get_temporal_high_water_mark_allocation_records(
                        merge_threads=merge_threads
                    )
                    reporter = FlameGraphReporter.from_temporal_snapshot(
                        recs,
                        memory_records=tuple(reader.get_memory_snapshots()),
                        native_traces=reader.metadata.has_native_traces,
                        high_water_mark_by_snapshot=hwms,
                        inverted=options.inverted,
                    )
            else:
                if options.show_memory_leaks:
                    snapshot = reader.get_leaked_allocation_records(
                        merge_threads=merge_threads
                    )
                elif options.temporary_allocation_threshold >= 0:
                    snapshot = reader.get_temporary_allocation_records(
                        threshold=options.temporary_allocation_threshold,
                        merge_threads=merge_threads,
                    )
                else:
                    snapshot = reader.get_high_watermark_allocation_records(
                        merge_threads=merge_threads
                    )

                memory_records = tuple(reader.get_memory_snapshots())
                reporter = FlameGraphReporter.from_snapshot(
                    snapshot,
                    memory_records=memory_records,
                    native_traces=options.native,
                    inverted=options.inverted,
                )

        assert reporter is not None
        flamegraph_path = tempdir / "flamegraph.html"
        with open(flamegraph_path, "w") as f:
            reporter.render(
                outfile=f,
                metadata=reader.metadata,
                show_memory_leaks=options.show_memory_leaks,
                merge_threads=merge_threads,
                inverted=options.inverted,
            )
        dump_file.unlink()
        pprint(f"Results saved to [bold cyan]{flamegraph_path}")
        _display_iframe(IFrame(str(flamegraph_path), width="100%", height="600"))


assert FlamegraphMagics.memray_flamegraph.__doc__ is not None
FlamegraphMagics.memray_flamegraph.__doc__ += "\n\n" + argument_parser().format_help()



import html
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import TextIO
from typing import Tuple

from memray import AllocationRecord
from memray import AllocatorType
from memray import MemorySnapshot
from memray import Metadata
from memray.reporters.common import format_thread_name
from memray.reporters.templates import render_report


class TableReporter:
    def __init__(
        self,
        data: List[Dict[str, Any]],
        *,
        memory_records: Iterable[MemorySnapshot],
    ):
        super().__init__()
        self.data = data
        self.memory_records = memory_records

    @classmethod
    def from_snapshot(
        cls,
        allocations: Iterable[AllocationRecord],
        *,
        memory_records: Iterable[MemorySnapshot],
        native_traces: bool,
        merge_threads: bool = True,
        **kwargs: Any,
    ) -> "TableReporter":
        # Aggregate records that share the same displayed fields, since the
        # table only shows the top stack frame and records with different full
        # call stacks can map to the same displayed row.
        aggregated: Dict[Tuple[str, str, str], Dict[str, Any]] = {}
        for record in allocations:
            stack_trace = (
                list(record.hybrid_stack_trace(max_stacks=1))
                if native_traces
                else record.stack_trace(max_stacks=1)
            )
            stack = "???"
            if stack_trace:
                function, file, line = stack_trace[0]
                stack = f"{function} at {file}:{line}"

            allocator = AllocatorType(record.allocator)
            tid = format_thread_name(record)
            # Note: When threads are merged the thread ID column must be left
            # out of the key, because the C++ aggregation currently leaves in
            # place the TID of one of the merged allocations.
            key = ("" if merge_threads else tid, allocator.name.lower(), stack)
            if key in aggregated:
                aggregated[key]["size"] += record.size
                aggregated[key]["n_allocations"] += record.n_allocations
            else:
                aggregated[key] = {
                    "tid": tid,
                    "size": record.size,
                    "allocator": allocator.name.lower(),
                    "n_allocations": record.n_allocations,
                    "stack_trace": html.escape(stack),
                }

        return cls(list(aggregated.values()), memory_records=memory_records)

    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        if inverted:
            raise NotImplementedError(
                "TableReporter does not support inverted argument"
            )
        html_code = render_report(
            kind="table",
            data=self.data,
            metadata=metadata,
            memory_records=self.memory_records,
            show_memory_leaks=show_memory_leaks,
            merge_threads=merge_threads,
            inverted=inverted,
            no_web=no_web,
        )
        print(html_code, file=outfile)



"""Utilities for extracting module names from file paths."""

import site
import sys
import sysconfig
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from typing_extensions import TypedDict


class PathInfo(TypedDict):
    stdlib: Optional[Path]
    site_packages: List[Path]
    sys_path: List[Path]


def get_python_path_info() -> PathInfo:
    """Get information about Python's search paths.

    Returns:
        dict: Dictionary containing stdlib path, site-packages paths, and sys.path entries.
    """
    libdest = sysconfig.get_config_var("LIBDEST")
    stdlib: Optional[Path] = Path(libdest) if libdest else None

    # Get site-packages directories
    site_packages: List[Path] = [Path(p) for p in site.getsitepackages()]

    # Get user site-packages
    user_site = site.getusersitepackages()
    if Path(user_site).exists():
        site_packages.append(Path(user_site))

    return {
        "stdlib": stdlib,
        "site_packages": site_packages,
        "sys_path": [Path(p) for p in sys.path if p],
    }


def _is_relative_to(path: Path, base: Path) -> bool:
    try:
        path.relative_to(base)
        return True
    except ValueError:
        return False


def _path_to_module(path: Path) -> str:
    if path.is_absolute():
        raise ValueError(f"Expected a relative path, got: {path}")

    if path.name == "__init__.py":
        path = path.parent
    elif path.suffix == ".py":
        path = path.with_suffix("")

    return ".".join(path.parts)


class ModuleResolver:
    """Resolve file paths to module names, caching results per file name.

    Holds the Python search path info and a cache so that the (relatively
    expensive) classification is only computed once per file name.
    """

    def __init__(self, path_info: Optional[PathInfo] = None) -> None:
        self._path_info: PathInfo = (
            path_info if path_info is not None else get_python_path_info()
        )
        self._cache: Dict[str, Tuple[str, str]] = {}

    def extract_module_name_and_type(self, filename: str) -> Tuple[str, str]:
        """Extract Python module name and type from a file path.

        Returns:
            tuple: (module_name, module_type) where module_type is one of:
                   'stdlib', 'site-packages', 'project', or 'unknown'
        """
        try:
            return self._cache[filename]
        except KeyError:
            result = self._extract_module_name_and_type_impl(filename)
            self._cache[filename] = result
            return result

    def _extract_module_name_and_type_impl(self, filename: str) -> Tuple[str, str]:
        if not filename:
            return ("unknown", "unknown")

        if filename.startswith("<frozen "):
            return (filename[len("<frozen ") : -1], "stdlib")

        file_path = Path(filename)

        for site_pkg in self._path_info["site_packages"]:
            if _is_relative_to(file_path, site_pkg):
                return (
                    _path_to_module(file_path.relative_to(site_pkg)),
                    "site-packages",
                )

        stdlib = self._path_info["stdlib"]
        if stdlib and _is_relative_to(file_path, stdlib):
            return (_path_to_module(file_path.relative_to(stdlib)), "stdlib")

        for path_entry in self._path_info["sys_path"]:
            if _is_relative_to(file_path, path_entry):
                return (_path_to_module(file_path.relative_to(path_entry)), "project")

        # Fallback: use just the filename, not the full absolute path
        return (_path_to_module(Path(file_path.name)), "unknown")

    def get_module_for_stack(self, stack: Iterable[Tuple[str, str, int]]) -> str:
        """Find the top-level module of the closest non-stdlib frame in a stack.

        Walks frames from leaf to root, returning the first non-stdlib module's
        top-level package name. Returns "<root>" if we reach Memray's own calls
        into the user's code, or if all frames are stdlib or the stack's empty.
        """
        for frame in stack:
            module_name, module_type = self.extract_module_name_and_type(frame[1])
            if module_type != "stdlib":
                top_level = module_name.split(".")[0]
                if top_level == "memray":
                    return "<root>"
                return top_level
        return "<root>"



"""Tools for processing and filtering stack frames."""

import contextlib
import functools
import linecache
import os
import re
from typing import Tuple

Symbol = str
File = str
Lineno = int
StackFrame = Tuple[Symbol, File, Lineno]

RE_CPYTHON_PATHS = re.compile(r"(Include|Objects|Modules|Python|cpython).*\.[c|h]$")

SYMBOL_IGNORELIST = {
    "PyObject_Call",
    "call_function",
    "classmethoddescr_call",
    "cmpwrapper_call",
    "do_call_core",
    "fast_function",
    "function_call",
    "function_code_fastcall",
    "instance_call",
    "instancemethod_call",
    "instancemethod_call",
    "methoddescr_call",
    "proxy_call",
    "slot_tp_call",
    "trace_call_function",
    "type_call",
    "weakref_call",
    "wrap_call",
    "wrapper_call",
    "wrapperdescr_call",
}


@functools.lru_cache(maxsize=1000)
def _is_cpython_internal_symbol(symbol: str, file: str) -> bool:
    if "PyEval_EvalFrameEx" in symbol or "_PyEval_EvalFrameDefault" in symbol:
        is_candidate = True
    elif symbol.startswith("_TAIL_CALL_") and ".llvm." in symbol:
        # Python 3.14 tail call interpreter uses LLVM-generated functions
        is_candidate = True
    elif symbol.startswith(("PyEval", "_Py")):
        is_candidate = True
    elif "vectorcall" in symbol.lower():
        is_candidate = True
    elif symbol in SYMBOL_IGNORELIST:
        is_candidate = True
    elif "Objects/call.c" in file:
        is_candidate = True
    else:
        is_candidate = False

    if is_candidate:
        return re.search(RE_CPYTHON_PATHS, file) is not None
    return False


def is_cpython_internal(frame: StackFrame) -> bool:
    symbol, file, _ = frame
    return _is_cpython_internal_symbol(symbol, file)


def is_frame_interesting(frame: StackFrame) -> bool:
    function, file, _ = frame

    if file.endswith("runpy.py") or file == "<frozen runpy>":
        return False

    return not _is_cpython_internal_symbol(function, file)


def is_frame_from_import_system(frame: StackFrame) -> bool:
    function, file, _ = frame
    if "frozen importlib" in file:
        return True
    if function in {"import_name", "import_from", "import_all_from"} and file.endswith(
        "ceval.c"
    ):
        return True
    return False


@functools.cache
def is_confidential_file(file: str) -> bool:
    """Attempt to guess whether the file is likely to contain secrets.

    We don't want to include secrets in flame graphs, since flame graphs may be
    shared with other people.

    We assume that Python source files won't contain keys, certificates, or
    other tokens, but other files might. We assume that a file with a `.py`
    extension is a Python source file, and that an executable file is likely to
    be a Python source file too. We assume that any world-readable file will
    not contain secrets.

    These heuristics may not always be correct, and flame graphs may wind up
    having secrets embedded in them. Don't share a Memray flame graph with
    someone unless you trust them.
    """
    if file.endswith(".py"):
        return False
    with contextlib.suppress(OSError):
        st = os.stat(file)
        if st.st_mode & 0o111 or st.st_mode & 0o004:
            return False
    return True


def source_line_from_frame(
    frame: StackFrame, confidential_files: str = "default"
) -> str:
    function, filename, lineno = frame

    if confidential_files == "all":
        is_confidential = True
    elif confidential_files == "none":
        is_confidential = False
    else:
        is_confidential = is_confidential_file(filename)

    if is_confidential:
        return f"{function} at {filename}:{lineno}"
    return linecache.getline(filename, lineno) or f"{function} at {filename}:{lineno}"



import argparse

from ..reporters.flamegraph import FlameGraphReporter
from .common import HighWatermarkCommand


class FlamegraphCommand(HighWatermarkCommand):
    """Generate an HTML flame graph for peak memory usage"""

    def __init__(self) -> None:
        super().__init__(
            reporter_factory=FlameGraphReporter.from_snapshot,
            temporal_reporter_factory=FlameGraphReporter.from_temporal_snapshot,
            reporter_name="flamegraph",
        )

    def prepare_parser(self, parser: argparse.ArgumentParser) -> None:
        super().prepare_parser(parser)
        parser.add_argument(
            "--split-threads",
            help="Do not merge allocations across threads",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--inverted",
            help="Invert flame graph",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--max-memory-records",
            help="Maximum number of memory records to display",
            type=int,
            default=None,
        )

        parser.add_argument(
            "--no-web",
            help="Use local assets instead of fetching from CDN",
            action="store_true",
            default=False,
        )

        parser.add_argument(
            "--confidential-files",
            help=(
                "Control what source code is included in the flame graph."
                " If set to 'all', all files are treated as confidential and"
                " no source code is included in the report (only function"
                " names, file names, and line numbers). If set to 'none',"
                " no files are treated as confidential and lines from any file"
                " may be included in the report. If set to 'default' (the"
                " default) Memray uses heuristics to guess whether each file"
                " is likely to contain secrets. It includes source lines from"
                " .py files and from any file that is world-readable or"
                " executable, and excludes lines from other files."
            ),
            choices=["all", "none", "default"],
            default="default",
        )



import csv
import json
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Sequence
from typing import TextIO
from typing import Tuple

from memray import AllocationRecord
from memray import AllocatorType
from memray import MemorySnapshot
from memray import Metadata
from memray._version import __version__
from memray.reporters.common import format_thread_name

Location = Tuple[str, str]
FrameLocation = Tuple[str, str, int]
FrameSample = Tuple[int, ...]


@dataclass
class _SampleWeights:
    size: int = 0
    n_allocations: int = 0


class TransformReporter:
    SUFFIX_MAP = {
        "gprof2dot": ".json",
        "csv": ".csv",
        "speedscope": ".speedscope.json",
    }

    def __init__(
        self,
        allocations: Iterable[AllocationRecord],
        *,
        format: str,
        native_traces: bool,
        memory_records: Iterable[MemorySnapshot],
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.allocations = allocations
        self.format = format
        self.native_traces = native_traces
        self.memory_records = memory_records

    def render_as_gprof2dot(
        self,
        outfile: TextIO,
        **kwargs: Any,
    ) -> None:
        location_to_index: Dict[Location, int] = {}
        all_locations: List[Dict[str, str]] = []
        events = []
        for record in self.allocations:
            stack_trace = self._stack_trace_for_record(record)
            call_chain = []
            for func, mod, _ in stack_trace:
                location = (func, mod)
                index = location_to_index.get(location)
                if index is None:
                    index = len(all_locations)
                    all_locations.append({"name": func, "module": mod})
                    location_to_index[location] = index
                call_chain.append(index)

            if not call_chain:
                continue
            events.append({"callchain": call_chain, "cost": [record.size]})

        result = {
            "version": 0,
            "costs": [{"description": "Memory", "unit": "bytes"}],
            "events": events,
            "functions": all_locations,
        }
        json.dump(result, outfile)

    def _stack_trace_for_record(
        self, record: AllocationRecord
    ) -> Sequence[Tuple[str, str, int]]:
        return (
            record.hybrid_stack_trace() if self.native_traces else record.stack_trace()
        )

    def _speedscope_sample_for_record(
        self,
        record: AllocationRecord,
        *,
        location_to_index: Dict[FrameLocation, int],
        frames: List[Dict[str, Any]],
    ) -> FrameSample:
        stack_trace = self._stack_trace_for_record(record)
        if not stack_trace:
            return ()

        # Speedscope sampled stacks are root-to-leaf. Memray exposes leaf-to-root.
        sample = []
        for func, mod, line in reversed(stack_trace):
            location = (func, mod, line)
            index = location_to_index.get(location)
            if index is None:
                index = len(frames)
                frame: Dict[str, Any] = {"name": func}
                if mod:
                    frame["file"] = mod
                if line > 0:
                    frame["line"] = line
                frames.append(frame)
                location_to_index[location] = index
            sample.append(index)
        return tuple(sample)

    def _aggregate_speedscope_samples(
        self,
        allocations: Iterable[AllocationRecord],
    ) -> Tuple[List[Dict[str, Any]], Dict[FrameSample, _SampleWeights]]:
        location_to_index: Dict[FrameLocation, int] = {}
        frames: List[Dict[str, Any]] = []
        sample_weights: Dict[FrameSample, _SampleWeights] = {}

        for record in allocations:
            sample = self._speedscope_sample_for_record(
                record,
                location_to_index=location_to_index,
                frames=frames,
            )
            if not sample:
                continue

            weights = sample_weights.get(sample)
            if weights is None:
                weights = sample_weights[sample] = _SampleWeights()
            weights.size += record.size
            weights.n_allocations += record.n_allocations

        return frames, sample_weights

    def _create_speedscope_profile(
        self,
        *,
        name: str,
        unit: str,
        weighted_samples: Iterable[Tuple[FrameSample, int]],
    ) -> Dict[str, Any]:
        samples: List[List[int]] = []
        weights: List[int] = []

        for sample, weight in weighted_samples:
            if weight <= 0:
                continue
            samples.append(list(sample))
            weights.append(weight)

        return {
            "type": "sampled",
            "name": name,
            "unit": unit,
            "startValue": 0,
            "endValue": sum(weights),
            "samples": samples,
            "weights": weights,
        }

    def render_as_speedscope(
        self,
        outfile: TextIO,
        **kwargs: Any,
    ) -> None:
        metadata = kwargs.get("metadata")
        frames, sample_weights = self._aggregate_speedscope_samples(self.allocations)

        result = {
            "$schema": "https://www.speedscope.app/file-format-schema.json",
            "shared": {"frames": frames},
            "profiles": [
                self._create_speedscope_profile(
                    name="Memory",
                    unit="bytes",
                    weighted_samples=(
                        (sample, weights.size)
                        for sample, weights in sample_weights.items()
                    ),
                ),
                self._create_speedscope_profile(
                    name="Allocations",
                    unit="none",
                    weighted_samples=(
                        (sample, weights.n_allocations)
                        for sample, weights in sample_weights.items()
                    ),
                ),
            ],
            "name": metadata.command_line if metadata is not None else "memray",
            "activeProfileIndex": 0,
            "exporter": f"memray@{__version__}",
        }
        json.dump(result, outfile)

    def render(
        self,
        outfile: TextIO,
        metadata: Metadata,
        show_memory_leaks: bool,
        merge_threads: bool,
        inverted: bool,
        no_web: bool = False,
    ) -> None:
        if not merge_threads:
            raise NotImplementedError("TransformReporter only supports merged threads.")
        if inverted:
            raise NotImplementedError(
                "TransformReporter does not support inverted argument."
            )
        renderer = getattr(self, f"render_as_{self.format}")
        renderer(outfile, metadata=metadata, show_memory_leaks=show_memory_leaks)

    def render_as_csv(
        self,
        outfile: TextIO,
        **kwargs: Any,
    ) -> None:
        writer = csv.writer(outfile)
        writer.writerow(
            [
                "allocator",
                "num_allocations",
                "size",
                "tid",
                "thread_name",
                "stack_trace",
            ]
        )
        for record in self.allocations:
            stack_trace = self._stack_trace_for_record(record)
            writer.writerow(
                [
                    AllocatorType(record.allocator).name,
                    record.n_allocations,
                    record.size,
                    record.tid,
                    format_thread_name(record),
                    "|".join(f"{func};{mod};{line}" for func, mod, line in stack_trace),
                ]
            )
